/*
Copyright (C) 2017-2018 Philipp Maierhoefer

This file is part of pyRed.

pyRed is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

pyRed is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with pyRed.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PYRED_COEFF_VEC_H
#define PYRED_COEFF_VEC_H

#include <iostream>
#include <string>
#include <vector>
#include <utility>
#include <mutex>
#include <type_traits>

#include "defs.h"

namespace pyred {

using cvec = std::vector<uint64_t>;

class Coeff_vec
{
  friend Coeff_vec operator-(const Coeff_vec&);
  friend Coeff_vec operator+(const Coeff_vec&, const Coeff_vec&);
  friend Coeff_vec operator-(const Coeff_vec&, const Coeff_vec&);
  friend Coeff_vec operator*(const Coeff_vec&, const Coeff_vec&);
  friend Coeff_vec operator/(const Coeff_vec&, const Coeff_vec&);
  friend Coeff_vec pow(Coeff_vec&&, const Coeff_vec&);
  friend bool operator==(const Coeff_vec&, const Coeff_vec&);
  friend bool operator!=(const Coeff_vec&, const Coeff_vec&);
  friend std::ostream& operator<<(std::ostream&, const Coeff_vec&);
private:
  cvec c;
  static uint32_t ncoeffs; // number of coefficients in coefficient vector
  static cvec prime;
  static bool auto_vars;
  static std::string set_to_one;
  static std::vector<std::pair<std::string,cvec>> randvars;
  static std::mutex s_mtx;
public:
  static void init();
  Coeff_vec();
  Coeff_vec(const Coeff_vec&);
  Coeff_vec(Coeff_vec&&);
  Coeff_vec(const std::string&);
  template<typename T,
           typename=typename std::enable_if<std::is_integral<T>::value>::type>
  Coeff_vec(T);
  Coeff_vec& operator=(const Coeff_vec&) = default;
  Coeff_vec& operator=(Coeff_vec&&) = default;
  Coeff_vec& operator+=(const Coeff_vec&);
  Coeff_vec& operator-=(const Coeff_vec&);
  Coeff_vec& operator*=(const Coeff_vec&);
  Coeff_vec& operator/=(const Coeff_vec&);
  operator bool() const;
  std::size_t hash() const {return c.front();}
};

Coeff_vec operator-(const Coeff_vec&);
Coeff_vec operator+(const Coeff_vec&, const Coeff_vec&);
Coeff_vec operator-(const Coeff_vec&, const Coeff_vec&);
Coeff_vec operator*(const Coeff_vec&, const Coeff_vec&);
Coeff_vec operator/(const Coeff_vec&, const Coeff_vec&);
Coeff_vec pow(Coeff_vec&&, const Coeff_vec&);
inline Coeff_vec pow(const Coeff_vec& base, const Coeff_vec& exp) {
  return pow(Coeff_vec{base}, exp);
}
bool operator==(const Coeff_vec&, const Coeff_vec&);
bool operator!=(const Coeff_vec&, const Coeff_vec&);
std::ostream& operator<<(std::ostream&, const Coeff_vec&);

template<typename T, typename>
inline Coeff_vec::Coeff_vec(T a): c(cvec()) {
  if (a >= 0) {
    c.resize(ncoeffs, a);
  }
  else {
    auto cc = uint64_t(-a);
    c.reserve(ncoeffs);
    for (const auto& p: prime) {
      c.emplace_back(p-cc);
    }
  }
}

inline Coeff_vec::operator bool() const {
  for (const auto& cc: c) {
    if (cc) return true;
  }
  return false;
}

}

namespace std {
  template<>
  struct hash<pyred::Coeff_vec>
  {
    size_t operator()(const pyred::Coeff_vec& cf) const {
      return cf.hash();
    }
  };
}

#endif
