/*
Copyright (C) 2017-2018 Philipp Maierhoefer

This file is part of pyRed.

pyRed is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

pyRed is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with pyRed.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <fstream>
#include <stack>
#include <stdexcept>
#include <unistd.h> // getopt

#include "config.h"
#include "interface.h"
#include "parser.h"

namespace pyred {

float time_diff(const std::clock_t& start, const std::clock_t& end) {
  return (float(end-start)/CLOCKS_PER_SEC);
}

float time_diff(const std::chrono::time_point<std::chrono::system_clock>& start,
                const std::chrono::time_point<std::chrono::system_clock>& end) {
  return static_cast<std::chrono::duration<float>>(end-start).count();
}

/***************************
 * Logger and LockedLogger *
 ***************************/

Logger::Logger(int verb): m_verbosity{verb} {}

void Logger::verbosity(int verb) {m_verbosity = verb;}

int Logger::verbosity() {return m_verbosity;}

LockedLogger Logger::operator()(int lev) {
  return LockedLogger(*this, lev);
}

LockedLogger::LockedLogger(Logger& lg, int lev)
: m_verbosity{lg.m_verbosity}, m_level{lev}, m_lck{lg.m_mtx}
{
  std::cout << std::flush;
}

LockedLogger& LockedLogger::operator<<(std::ostream& (*omanip)(std::ostream&)) {
  if (m_level <= m_verbosity) std::cout << (*omanip);
  return *this;
}

/**********
 * Config *
 **********/

Logger Config::s_log_val{1};
int Config::s_coeff_cls_val{1};
bool Config::s_backward_val{true};
int Config::s_parallel{1};
bool Config::s_auto_symseed{true};
std::pair<int,int> Config::s_symlimits{std::numeric_limits<int>::max(),2};
int Config::s_lookahead{1};
int Config::s_insertion_tracer{1};
std::pair<std::string,bool> Config::s_database_file{"./insertions",false};
bool Config::s_finished{false};

void Config::coeff_cls(int ccls) {
  if (ccls != 1 && ccls != 2) {
    std::ostringstream ss;
    ss << "Invalid coefficient class number (must be 1 or 2): " << ccls;
    throw init_error(ss.str());
  }
  s_coeff_cls_val = ccls;
}

void Config::backward(bool bkwd) {
  s_backward_val = bkwd;
}

void Config::verbosity(int verb) {
  s_log_val.verbosity(verb);
}

void Config::parallel(int n) {
  if (n < 0) n = 1;
  s_parallel = n;
}

void Config::auto_symseed(const bool ass) {
  s_auto_symseed = ass;
}

void Config::symlimits(const int maxdots, const int maxsps) {
  s_symlimits = {maxdots, maxsps};
}

void Config::lookahead(int lah) {
  if (lah < -1 || lah > 2) {
    throw init_error("Config::lookahead may only be initialised to -1,0,1,2.");
  }
  s_lookahead = lah;
}

void Config::insertion_tracer(int mode) {
  if (mode < 0 || mode > 4) {
    throw init_error("Config::insertion_tracer may only be initialised to "
                     "0 (off), 1 (in memory, default), 2 (SQLite), "
                     "3 (Kyoto Cabinet), "
                     "4 (Kyoto Cabinet if available, otherwise SQLite).");
  }
#ifdef PYRED_KCDB
  if (mode == 4) mode = 3;
#else
  if (mode == 4) mode = 2;
  if (mode == 3) {
    throw init_error("To use Config::insertion_tracer=3, pyRed must be built "
                     "with Kyoto Cabinet database support.");
  }
#endif
  s_insertion_tracer = mode;
}

std::pair<std::string,std::string> Config::parse(const std::string& conf) {
  auto tmp = split(conf, ' ');
  tmp.insert(tmp.begin(), ""); // getopt() ignores the first element
  return parse(tmp);
}

std::pair<std::string,std::string> Config::parse(
    const std::vector<std::string>& conf) {
  if (s_finished) {
    throw init_error(
      "Config::parse() error: finalised config cannot be changed anymore.");
  }
  // convert conf to char**
  int argc = conf.size();
  std::vector<char*> argv;
  argv.reserve(argc);
  for (const auto& opt: conf) {
    argv.push_back(const_cast<char*>(opt.c_str()));
  }
  int opt;
  std::string infile{};
  std::string outfile{};
  optind = 1;
  while ((opt = getopt(argc, argv.data(), "1:c:i:o:s:b:t:v:")) != -1) {
    /*
     * -1:string symbol to set to one
     * -c:int use coefficient array of this size
     * -i:string declare symbol (comma separated or multiple -i options)
     * -o:string output file name
     * -s:int integer seed for random number generator
     * -b:int 0/1 to turn off/on backward insertion (default=1)
     */
    int seed;
    switch (opt) {
    case '1':
      CoeffHelper::settoone(std::string(optarg));
      break;
    case 'c':
      try {
        CoeffHelper::coeff_n(string_to_int(std::string(optarg)));
      }
      catch(std::invalid_argument& e) {
        std::ostringstream ss;
        ss << "Option c argument must be integer, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      coeff_cls(2);
      break;
    case 'i':
      for (const auto& var: split(std::string(optarg), ',')) {
        CoeffHelper::add_invariant(var);
      }
      break;
    case 'o':
      outfile = optarg;
      break;
    case 's':
      try {
        seed = string_to_int(std::string(optarg));
      }
      catch(std::invalid_argument& e) {
        std::ostringstream ss;
        ss << "Seed argument must be integer, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      CoeffHelper::random_seed(seed);
      break;
    case 'b':
      if (std::string(optarg) == "0") {
        backward(false);
      }
      else if (std::string(optarg) == "1") {
        backward(true);
      }
      else {
        std::ostringstream ss;
        ss << "Option -b must either be 0 or 1, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      break;
    case 'v':
      try {
        verbosity(string_to_int(std::string(optarg)));
      }
      catch(std::invalid_argument& e) {
        std::ostringstream ss;
        ss << "Verbosity argument must be integer, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      break;
    default:
      std::ostringstream ss;
      ss << "Invalid argument " << opt << std::endl;
      throw init_error(ss.str());
    }
  }
  if (optind == argc-1) {
    infile = argv[optind];
  }
  else if (optind < argc-1) {
    std::ostringstream ss;
    ss << "Too many arguments" << std::endl;
    throw init_error(ss.str());
  }
  return {infile, outfile};
}

void Config::finish() {
  if (coeff_cls() == 1) {
    Coeff_int::init();
  }
  else {
    Coeff_vec::init();
  }
  s_finished = true;
}

/**********
 * System *
 **********/

System::System()
: sys{}, coeff_cls{Config::coeff_cls()}, m_content_prepared{false}
{}

System::System(const std::vector<eqdata>& eqs)
: sys{eqs}, coeff_cls{Config::coeff_cls()}, m_content_prepared{false}
{}

System::System(std::vector<eqdata>&& eqs)
: sys{std::move(eqs)}, coeff_cls{Config::coeff_cls()}, m_content_prepared{false}
{}

System::System(const std::string& infile)
: sys{}, coeff_cls{Config::coeff_cls()}, m_content_prepared{false}
{
  add(infile);
}

std::size_t System::size() const {
  if (systemtype() == 0) {
    return sys.size();
  }
  else {
    // at most one can be non-empty
    return numsys_int.size() + numsys_vec.size();
  }
}

int System::systemtype(bool fatal_empty) const {
  if (!sys.empty() && numsys_int.empty() && numsys_vec.empty()) {
    // String coefficients.
    return 0;
  }
  else if (sys.empty() && ((!numsys_int.empty() &&  numsys_vec.empty()) ||
                           ( numsys_int.empty() && !numsys_vec.empty())) ) {
    // Finite integer coefficients of type (1) Coeff_int or (2) Coeff_vec.
    return coeff_cls;
  }
  else if (sys.empty() && numsys_int.empty() && numsys_vec.empty()) {
    if (fatal_empty) {
      throw init_error("System is empty.");
    }
    else {
      // Empty system.
      return -1;
    }
  }
  else {
    throw init_error("System mixes different coefficient types.");
  }
}

std::vector<intid> System::generate_solve(
    const std::vector<SeedSpec>& ibp_seedspec,
    const std::vector<SeedSpec>& sym_seedspec) {
  if (Integral::sector_ordering() == 1 &&
      (Config::lookahead() == 0 || Config::lookahead() == 1)) {
    // In sector ordering 1, sectors are not ordered by number of lines.
    // Therefore the on-the-fly generator/solver which sorts equations
    // across seed sectors (i.e. lookahead != -1) can only be used with
    // lookahead=2, i.e. the entire system is generated before it is solved.
    throw init_error("System::generate_solve() with sectors ordering 1 "
      "(i.e. by sector number) can only be used with lookahead=-1 "
      "(no sorting across seed sectors) or lookahead=2 "
      "(generate the entire system before sorting).");
  }
  m_is_generated = true;
  m_seeds = std::make_pair(ibp_seedspec, sym_seedspec);
  std::vector<intid> indep_eqnums;
  if (coeff_cls == 1) {
    indep_eqnums = generate_solve_tmpl<Coeff_int>(ibp_seedspec, sym_seedspec);
  }
  else {
    indep_eqnums = generate_solve_tmpl<Coeff_vec>(ibp_seedspec, sym_seedspec);
  }
  if (Config::backward()) {
    backward();
  }
  return indep_eqnums;
}

void System::generate_retrieve(
    const std::vector<SeedSpec>& ibp_seedspec,
    const std::vector<SeedSpec>& sym_seedspec,
    const std::function<void(eqdata &&)> &treateq,
    const std::function<std::string(const std::string&)> &treatcoeff) {
  // Generate a system of equation for the given seed, insert coefficients
  // as strings, optionally simplified by treatcoeff,
  // and apply treateq to all equations (i.e. no selection).
  if (coeff_cls == 1) {
    GeneratorHelper::generate_and_retrieve<Coeff_int>(
      ibp_seedspec,
      sym_seedspec,
      Config::auto_symseed(),
      {},
      treatcoeff,
      treateq,
      Config::parallel());
  }
  else {
    GeneratorHelper::generate_and_retrieve<Coeff_vec>(
      ibp_seedspec,
      sym_seedspec,
      Config::auto_symseed(),
      {},
      treatcoeff,
      treateq,
      Config::parallel());
  }
}

void System::generate_retrieve(
    std::vector<intid> &&eqnums,
    const std::function<void(eqdata &&)> &treateq,
    const std::function<std::string(const std::string&)> &treatcoeff) {
  // Re-generate the system, insert coefficients as strings,
  // optionally simplified by treatcoeff,
  // and apply treateq to the (selected) equations with numbers in eqnums.
  if (!m_is_generated) {
    throw init_error(
      "System::generate_retrieve(): seeds have not been set for re-generation."
    );
  }
  if (eqnums.empty()) {
    return;
  }
  if (coeff_cls == 1) {
    GeneratorHelper::generate_and_retrieve<Coeff_int>(
      m_seeds.first,
      m_seeds.second,
      Config::auto_symseed(),
      std::move(eqnums),
      treatcoeff,
      treateq,
      Config::parallel());
  }
  else {
    GeneratorHelper::generate_and_retrieve<Coeff_vec>(
      m_seeds.first,
      m_seeds.second,
      Config::auto_symseed(),
      std::move(eqnums),
      treatcoeff,
      treateq,
      Config::parallel());
  }
}

void System::reserve(const std::size_t sz) {
  sys.reserve(sz);
}

void System::add(const eqdata& eq) {
  sys.push_back(eq);
}

void System::add(eqdata&& eq) {
  sys.push_back(std::move(eq));
}

void System::add(const std::vector<eqdata>& eqs){
  reserve(sys.size() + eqs.size());
  for (const auto& eq: eqs) {
    sys.push_back(eq);
  }
}

void System::add(std::vector<eqdata>&& eqs){
  reserve(sys.size() + eqs.size());
  for (auto& eq: eqs) {
    sys.push_back(std::move(eq));
  }
}

void System::add_forward(const eqdata& eq, intid neq) {
  if (coeff_cls == 1) {
    add_forward_tmpl<Coeff_int>(eq, get_numsys<Coeff_int>(), sols_int, neq);
  }
  else {
    add_forward_tmpl<Coeff_vec>(eq, get_numsys<Coeff_vec>(), sols_vec, neq);
  }
}

void System::add(const std::string& infile) {
  Config::log(1) << "import equations" << std::flush;
  auto starttime = std::clock();
  std::string line;
  eqdata tmpeq;
  if (infile.empty()) {
    // read equations from stdin
    while (std::getline(std::cin, line)) {
      if (line.empty() && !tmpeq.empty()) {
        sys.push_back(std::move(tmpeq));
        tmpeq.clear();
      }
      else {
        auto wc = split_intcoeff(line);
        if (wc.first != Integral::zero_weight) {
          tmpeq.push_back(std::move(wc));
        }
      }
    }
    if (!tmpeq.empty()) {
      sys.push_back(std::move(tmpeq));
    }
  }
  else {
    // import equations from file
    std::ifstream instream(infile);
    if (!instream.good()) {
      std::ostringstream ss;
      ss << "failed reading file \"" << infile << "\"" << std::endl;
      throw input_error(ss.str());
    }
    while (std::getline(instream, line)) {
      if (line.empty() && !tmpeq.empty()) {
        sys.push_back(std::move(tmpeq));
        tmpeq.clear();
      }
      else {
        auto wc = split_intcoeff(line);
        if (wc.first != Integral::zero_weight) {
          tmpeq.push_back(std::move(wc));
        }
      }
    }
  }
  if (!tmpeq.empty()) {
    sys.push_back(std::move(tmpeq));
  }
  auto endtime = std::clock();
  Config::log(1) << ":   " << time_diff(starttime, endtime) << std::endl;
}

std::vector<intid> System::solve() {
  std::vector<intid> indep_eqnums;
  if (coeff_cls == 1) {
    indep_eqnums = solve_tmpl<Coeff_int>();
  }
  else {
    indep_eqnums = solve_tmpl<Coeff_vec>();
  }
  return indep_eqnums;
}

void System::backward() {
  if (coeff_cls == 1) {
    sols_int.clear();
    backward_tmpl<Coeff_int>(get_numsys<Coeff_int>());
  }
  else {
    sols_vec.clear();
    backward_tmpl<Coeff_vec>(get_numsys<Coeff_vec>());
  }
}

std::vector<intid> System::independent() {
  std::vector<intid> indep_eqnums;
  std::size_t nequations;
  if (coeff_cls == 1) {
    auto &numsys = get_numsys<Coeff_int>();
    nequations = numsys.size();
    indep_eqnums = numsys.independent();
  }
  else {
    auto &numsys = get_numsys<Coeff_vec>();
    nequations = numsys.size();
    indep_eqnums = numsys.independent();
  }
  auto neqs_indep = indep_eqnums.size();
  Config::log(1) << std::setw(8) << nequations << " equations: "
                 << nequations-neqs_indep << " zero + "
                 << neqs_indep << " independent" << std::endl;
  return indep_eqnums;
}

std::unique_ptr<keyvaluedb::KeyValueDB> &System::get_db() {
  if (coeff_cls == 1) {
    return get_numsys<Coeff_int>().get_db();
  }
  else {
    return get_numsys<Coeff_vec>().get_db();
  }
}

std::pair<std::vector<intid>,std::vector<intid>> System::select(
  const std::vector<intid>& mandatory, const std::vector<intid>& if_reduced) {
  /*
   * 'mandatory':
   * - Vector of integrals for which a reduction formula is required.
   * 'if_reduced':
   * - Vector of integrals to reduce, if there is a reduction formula known.
   * return: pair(neededeqs, unreduced)
   * 'neededeqs'
   * - The numbers of all equations which are needed to reduce integrals
   *   from 'mandatory' and 'if_reduced'.
   * 'unreduced':
   * - All integrals which appear in reduction formulas of mandatory integrals.
   * - All mandatory integrals for which there is no reduction formula known.
   */
  if (!m_content_prepared) {
    throw init_error("Content map has not been prepared for selection. "
                     "The system must be solved with backward insertion.");
  }
  // Stack of untreated integrals for which the dependencies must be looked up.
  std::stack<intid,std::vector<intid>> intstack;
  std::unordered_set<intid> neededset;
  std::unordered_set<intid> mandatoryset;
  // Set of integrals which are regarded as master integrals.
  std::unordered_set<intid> unreducedset;
  for (const auto iid: mandatory) {
    if (iid) neededset.insert(iid);
  }
  mandatoryset = neededset;
  for (const auto iid: if_reduced) {
    if (iid) neededset.insert(iid);
  }
  for (const auto iid: neededset) intstack.push(iid);
  while (!intstack.empty()) {
    auto next = intstack.top();
    intstack.pop();
    const auto cntit = m_content.find(next);
    if (cntit != m_content.cend()) {
      // A reduction formula is known.
      // Depends on equations with the numbers in the respective 'insertions'.
      // Get them from the insertions database.
      auto deps = get_db()->get(cntit->second.first/*eqnum*/, false);
      for (const auto& dep: deps) {
        // For each integral on which 'next' depends:
        // add it to the set of needed integrals.
        // If it was not already in the set, push it onto the stack.
        auto isnew = neededset.insert(dep).second;
        if (isnew) intstack.push(dep);
      }
      if (mandatoryset.find(next) != mandatoryset.cend()) {
        // If 'next' is in 'mandatory', treat integral content masters.
        for (const auto& iid: cntit->second.second/*unreduced*/) {
          unreducedset.insert(iid);
        }
      }
    }
    else {
      // No reduction formula is known.
      // Dependencies always correspond to (inserted) equations,
      // i.e. this branch is only reached for unreduced integrals
      // in 'mandatory' and 'if_reduced'.
      if (mandatoryset.find(next) != mandatoryset.cend()) {
        // 'next' is in 'mandatory', but there is no reduction formula.
        // -> Regard it as a master integral.
        unreducedset.insert(next);
      }
      else {
        // 'next' is in 'if_reduced', but there is no reduction formula.
        // -> Remove it from neededset. It won't be added again,
        //    because it was never inserted anywhere.
        // Do not regard it as a master integral.
        neededset.erase(next);
      }
    }
  }
  // For each integral id in 'neededset':
  // push the corresponding equation number on 'neededeqs',
  // if a reduction formula exists.
  std::vector<intid> neededeqs;
  neededeqs.reserve(neededset.size());
  for (const auto iid: neededset) {
    const auto cntit = m_content.find(iid);
    if (cntit != m_content.cend()) {
      neededeqs.push_back(cntit->second.first/*eqnum*/);
    }
  }
  std::sort(neededeqs.begin(), neededeqs.end());
  // Convert 'unreducedset' to a vector.
  std::vector<intid> unreduced;
  for (const auto iid: unreducedset) unreduced.push_back(iid);
  std::sort(unreduced.begin(), unreduced.end());
  return {neededeqs, unreduced};
}

std::vector<eqdata> System::retrieve(std::vector<intid> &&eqnums) {
  if (sys.empty()) {
    throw init_error("System::retrieve(): empty system.");
  }
  std::sort(eqnums.begin(), eqnums.end());
  std::vector<eqdata> result;
  intid neq = 0;
  auto selit = eqnums.begin();
  for (auto &eq: sys) {
    if (neq == *selit) {
      result.push_back(std::move(eq));
      ++selit;
    }
    eq.clear();
    ++neq;
  }
  sys.clear();
  return result;
}

}
