/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/

//jule is the number of propagators
#include "kira.h"
#include <algorithm>
#include <sys/stat.h>
#include "tools.h"
#include <tuple>

using namespace std;
using namespace GiNaC;

static Loginfo & logger = Loginfo::instance();

bool dsc(pair<unsigned,ex> i, pair<unsigned,ex> j) {return (i.first>j.first);};

bool dsc2(tuple<unsigned,unsigned,ex> i, tuple<unsigned,unsigned,ex> j) {return (get<0>(i) > get<0>(j));};


class LeeCriterium{
public:
  LeeCriterium(ex listOfTerms,Kira& kira, int degree);
  ~LeeCriterium();
  void prepare();
  void print();
  void collect();
  void getList(unsigned sector, lst& EQ);
  
  unsigned nOfPolynoms;
  unsigned nLoop;
  unsigned l_Indices;
  int *****FF;
  std::tuple<unsigned,unsigned,ex> ***fSum;
  unsigned **fsize;
  lst *listOfTerms;
  ex * holdTerm;
  int *a;
  possymbol *unknown;
  possymbol* feynman;
  ex * invarV;
  tuple<unsigned,unsigned,ex>* allTerms;
  unsigned nTerms;
};

void LeeCriterium::prepare(){
  
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for (unsigned j = 0; j < listOfTerms[i].nops(); j++) {
      for(unsigned k = 0; k < fsize[i][j]; k++){
	get<0>(fSum[i][j][k])=0;
        get<1>(fSum[i][j][k])=0;
	for(unsigned l = 0; l < nLoop; l++){
	  if(FF[i][j][k][l][0]==0){
	    get<0>(fSum[i][j][k])=0;
	    break;
	  }
	  get<0>(fSum[i][j][k])|=FF[i][j][k][l][0];
          get<1>(fSum[i][j][k])+=FF[i][j][k][l][0];
	}
      }
    }
  }
  
}

void LeeCriterium::print() {
  
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for (unsigned j = 0; j < listOfTerms[i].nops(); j++){
      
      for (unsigned k = 0; k < fsize[i][j]; k++){
	logger << get<0>(fSum[i][j][k]) << " ";
        logger << get<1>(fSum[i][j][k]) << " ";
	logger << get<2>(fSum[i][j][k]) << " ";
      }
      logger << "\n";
    }
  }
  
}

void LeeCriterium::collect() {

  nTerms=0;
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for (unsigned j = 0; j < listOfTerms[i].nops(); j++){
      nTerms+=fsize[i][j];
    }
  }

  allTerms = new tuple<unsigned,unsigned,ex> [nTerms];
  unsigned kk=0;
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for (unsigned j = 0; j < listOfTerms[i].nops(); j++){
      for (unsigned k = 0; k < fsize[i][j]; k++){
	allTerms[kk++]=fSum[i][j][k];
      }
    }
  }
  
  for(unsigned i = 0; i < nTerms; i++ ) {
    for(unsigned j = i+1; j < nTerms; j++ ) {
      if( get<0>(allTerms[i]) == get<0>(allTerms[j]) && get<1>(allTerms[i]) == get<1>(allTerms[j]) ) {
	get<2>(allTerms[i])+=get<2>(allTerms[j]);
	swap(allTerms[j], allTerms[nTerms-1]);
	nTerms--;
	j--;
      }
    }
  }
  
  sort(allTerms,allTerms+nTerms,dsc2);
    
  for(unsigned i = 0; i < nTerms; i++ ) {
    get<2>(allTerms[i]) = get<2>(allTerms[i]) == 0;
  }
  
//   for(unsigned i = 0; i < nTerms; i++ ) {
//     cout << get<2>(allTerms[i]) << endl;
//   }
  
}

void LeeCriterium::getList(unsigned sector, lst& EQ){
  
//   cout << "sector" << sector << endl;
  for(unsigned i = 0; i < nTerms; i++ ) {
//     cout << get<0>(allTerms[i]) << endl;
    if( ( get<0>(allTerms[i]) & sector ) == get<0>(allTerms[i]) ){
      EQ.append(get<2>(allTerms[i]));
    }
  }
}

LeeCriterium::LeeCriterium(ex polynom,Kira& kira, int degree){
  
  nLoop = kira.integralfamily.loopVar.size()+degree;
  nOfPolynoms=kira.invar.size()+1;
  l_Indices = kira.integralfamily.jule;
  
  ex kriterium;
  kriterium=polynom;
  
  a = new int [l_Indices];
  
  for(unsigned i = 0; i < l_Indices; i++)
    a[i]=(1<<i);
  
  listOfTerms = new lst [nOfPolynoms];
  holdTerm = new ex [l_Indices+1];
  unknown = new possymbol [l_Indices];
  
  generate_symbols(unknown,"k",l_Indices);
    
  ex holdPolynom;
  invarV = new ex [nOfPolynoms];
  for(unsigned i = 0; i < kira.invar.size(); i++) {

    for (int j = -kriterium.degree(1/kira.invar[i]); j <= kriterium.degree(kira.invar[i]); ++j) {

      if(j!=0 && kriterium.coeff(kira.invar[i],j)!=0){
        invarV[i]=(pow(kira.invar[i],j));
	holdPolynom=kriterium.coeff(kira.invar[i],j);
	for(unsigned k = 0; k < l_Indices; k++) {
	  holdTerm[k] = holdPolynom.coeff(unknown[k],1);
	  holdPolynom=holdPolynom.subs(unknown[k]==0);
	  listOfTerms[i].append(holdTerm[k]);
	}
	holdTerm[l_Indices] = holdPolynom;
	listOfTerms[i].append(holdTerm[l_Indices]);
        kriterium = expand(kriterium - kriterium.coeff(kira.invar[i],j)*(pow(kira.invar[i],j)));
      }
    }
  }
  
  if(kriterium!=0){
    holdPolynom=kriterium;
    for(unsigned k = 0; k < l_Indices; k++){
	holdTerm[k] = holdPolynom.coeff(unknown[k],1);
	holdPolynom=holdPolynom.subs(unknown[k]==0);
	listOfTerms[kira.invar.size()].append(holdTerm[k]);
    }
    holdTerm[l_Indices] = holdPolynom;
    listOfTerms[kira.invar.size()].append(holdTerm[l_Indices]);
    invarV[kira.invar.size()]=1;
  }
  
  FF = new int**** [nOfPolynoms];
  fSum = new tuple<unsigned,unsigned,ex>** [nOfPolynoms];
  fsize = new unsigned* [nOfPolynoms];
  
  for(unsigned i = 0; i < nOfPolynoms; i++) {
    
    FF[i] = new int*** [listOfTerms[i].nops()];
    fSum[i] = new tuple<unsigned,unsigned,ex> * [listOfTerms[i].nops()];
    fsize[i] = new unsigned [listOfTerms[i].nops()];
  }
  
  for (unsigned i = 0; i < nOfPolynoms; i++) {
    
    for (unsigned j = 0; j < listOfTerms[i].nops(); j++) {
      
      if(is_a<add>(listOfTerms[i].op(j))){
	
        FF[i][j] = new int** [listOfTerms[i].op(j).nops()];
	fSum[i][j] = new tuple<unsigned,unsigned,ex> [listOfTerms[i].op(j).nops()];
	fsize[i][j] = listOfTerms[i].op(j).nops();
      }
      else{
	
        FF[i][j] = new int** [1];
	fSum[i][j] = new tuple<unsigned,unsigned,ex> [1];
	fsize[i][j] = 1;
      }
      for (unsigned k = 0; k < fsize[i][j]; k++) {
	
        FF[i][j][k] = new int * [nLoop];
      }
    }
  }
  
  feynman = new possymbol [l_Indices];
  generate_symbols(feynman,"b",l_Indices);
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for (unsigned j = 0; j < listOfTerms[i].nops(); j++) {
      if(listOfTerms[i].op(j)==0){
	get<2>(fSum[i][j][0])=0;
	for(unsigned kk = 0; kk < nLoop; kk++){
	  FF[i][j][0][kk] = new int [1];
	  FF[i][j][0][kk][0]=0;
	}
	continue;
      }
      for(unsigned k = 0; k < fsize[i][j]; k++){
	
	int kk=0;
	ex term;
	
        if( fsize[i][j] > 1 )
          term = listOfTerms[i].op(j).op(k);
         
	else if ( fsize[i][j] == 1 )
	  term = listOfTerms[i].op(j);
        
        vector<ex> v_terms;
        unsigned l_terms;
        
        
        if((term.nops())) {
          l_terms=term.nops();
          for(unsigned l = 0; l < l_terms; l++){
            v_terms.push_back(term.op(l));
          }
        }
        else {
          l_terms = 1;
          v_terms.push_back(term);
        }
        
	
	get<2>(fSum[i][j][k])=invarV[i];
	
	if(j<l_Indices)
	  get<2>(fSum[i][j][k])*=unknown[j];
        
        if( is_a<numeric>(v_terms[l_terms-1])) {
	  get<2>(fSum[i][j][k])*=v_terms[l_terms-1];
        }
        
	for(unsigned l = 0; l < l_terms; l++){
	  if(l>nLoop)
	    break;
	  if(is_a<power>(v_terms[l])) {
	    exmap rpls;
	    (v_terms[l]).match(pow(wild(0),wild(1)),rpls);
	    for(unsigned m = 0; m < l_Indices; m++){
	      if(feynman[m]==rpls[wild(0)]){
		for(int ri = 0; ri < ex_to<numeric>(rpls[wild(1)]).to_int(); ri++){
		  FF[i][j][k][kk++]=&a[m];
		}
	      }
	    }
	  }
	  else {
	    for(unsigned m = 0; m < l_Indices; m++){
	      if(feynman[m]==v_terms[l]){
		FF[i][j][k][kk++]=&a[m];
	      }
	    }
	  }
	}
      }
    }  
  }
  prepare();
  collect();
  
};

LeeCriterium::~LeeCriterium(){
  
  for(unsigned i = 0; i < nOfPolynoms; i++) {
    for (unsigned j = 0; j < listOfTerms[i].nops(); j++) {
      for (unsigned k = 0; k < fsize[i][j]; k++) {
	if(get<2>(fSum[i][j][k]) == 0){
	  for(unsigned l = 0; l < nLoop; l++){
	    delete [] FF[i][j][k][l];
	  }
	}
	delete [] FF[i][j][k];
      }
      delete [] FF[i][j];
      delete [] fSum[i][j];
    }
    delete [] FF[i];
    delete [] fSum[i];
    delete [] fsize[i];
  }
  delete [] FF;
  delete [] fSum;
  delete [] fsize;
  
  delete [] listOfTerms;
  delete [] holdTerm;
  delete [] unknown;
  delete [] feynman;
  delete [] invarV;
  delete [] a;
  delete [] allTerms;
  
};


int Kira::find_zero_sectors() {
  
  logger << "\n*****Search trivial sectors********************************\n";
  Clock clock;
  
  int numberOfNonTrivialSectors = 0;
  
  int nn = integralfamily.loopVar.size();
  ex *vec, Skalar;
  vec = new ex [nn];
  possymbol token("token");
  
  /*generate Feynman parameters*/
  
  possymbol* feynman = new possymbol [integralfamily.jule];
  generate_symbols(feynman,"b",integralfamily.jule);

  ex schwinger;
  for (int i = 0; i < integralfamily.jule; i++){
    schwinger += -feynman[i]*integralfamily.props[i].expand();
  }
  fs<lst>(schwinger,kinematic);

  /*Create mZS, collect coefficients for loop_mom O(2)*/
  lst mZS;
  for (int i = 0; i < nn; i++){
    ex superSchwinger = schwinger;
    for (int j = 0; j < nn; j++){
      fs<relational>(superSchwinger,(integralfamily.loopVar[i]*integralfamily.loopVar[j] == token));
      if (i == j){
        mZS.append(-superSchwinger.coeff(token,1));
      }
      else{
        mZS.append(-superSchwinger.coeff(token,1)/2);
      }
      superSchwinger=superSchwinger.subs(token==0);
    }
  }

  for (int i = 0; i < nn; i++){
    for (int j = i; j < nn; j++){
      fs<relational>(schwinger,(integralfamily.loopVar[i]*integralfamily.loopVar[j] == token));
      fs<relational>(schwinger,token == 0);
    }
  }
  
  matrix mZS2(integralfamily.loopVar.size(), integralfamily.loopVar.size(),mZS);
  matrix mZS2inverse=mZS2.inverse();
  ex mZS2det=mZS2.determinant();

  /*Create vector, collect coefficients for loop_mom O(1)*/
  for (int i = 0; i < nn; i++) {
    fs<relational>(schwinger,integralfamily.loopVar[i] == token);
    vec[i] = schwinger.coeff(token,1)/2;
    fs<relational>(schwinger,token == 0);
  }
  ex J = schwinger;
  integralfamily.U = mZS2det.expand();
  
  /*calculate F*/
  ex ZScalar = 0,ZScalar2=0;
  for (int i = 0; i < nn; i++) {
    for (int j = 0; j < nn; j++){
      ZScalar += mZS2inverse(i,j)*vec[j]*vec[i];
    }
  }
  
  ex F = normal((ZScalar + J)*mZS2det).expand();
  fs<lst>(F,kinematic);
  F = expand(F.subs(kinematicReverse));
  
  integralfamily.F = F;
  
  
//   logger << integralfamily.U << "\n\n";
//   logger << integralfamily.F << "\n";
  
  integralfamily.G = integralfamily.U + integralfamily.F;
  
  ifstream input;
  string inputName = outputDir+"/sectormappings/"+integralfamily.name+"/nonTrivialSector";
  if(file_exists(inputName.c_str())){
    input.open(inputName.c_str());
    logger << "read from: " << inputName << "\n";
    while(1){
      
      int sector;
      int nOfProps;
      if(!(input >>sector)) break;
      if(!(input >>nOfProps)) break;
      numberOfNonTrivialSectors++;
      integralfamily.mask[nOfProps].push_back(sector);
    }
    input.close();
  
    logger << "\nNon trivial sectors in total: " << numberOfNonTrivialSectors << "\n";
    logger << "Trivial sectors in total: " << ((1<<integralfamily.jule)-numberOfNonTrivialSectors)<<"\n";
    logger << "( " << clock.eval_time() << " s )\n\n";
    
    delete [] vec;
    delete [] feynman;
  
    if(numberOfNonTrivialSectors == 0){
      logger << "The system you try to reduce contains only zero integrals.\n";
      return 1;
    }
    else
      return 0;
  }
   
  /*initialize Lee's zero sector criterium*/
  possymbol *unknown = new possymbol [integralfamily.jule];
  generate_symbols(unknown,"k",integralfamily.jule);
  ex kriterium = 0, kriterium2 = 0, kritU=0, kritF=0;
  for (int i = 0; i < integralfamily.jule; i++) {
    kriterium += unknown[i]*feynman[i]*integralfamily.G.diff(feynman[i]);
    kritU += unknown[i]*feynman[i]*integralfamily.U.diff(feynman[i]);
    kritF += unknown[i]*feynman[i]*integralfamily.F.diff(feynman[i]);
  }
  kriterium = expand(kriterium - integralfamily.G);
  kritU = expand(kritU - integralfamily.U);
  kritF = expand(kritF - integralfamily.F);

  lst exp_list, exp_list2, exp_list3, EQ, EQ2, comparefeynman;
  possymbol *exponent = new possymbol [integralfamily.jule];
  generate_symbols(exponent,"e",integralfamily.jule);

  ex monom = 1;
  for(int i = 0; i < integralfamily.jule; i++ ) {
    comparefeynman.append(feynman[i] == 0);
    monom *= GiNaC::pow(feynman[i],exponent[i]);
  }
    
//   cout << kritU << endl<< endl;
//   cout << kritF << endl;
  
  LeeCriterium hallo(kritF,*this,1);      
  
  LeeCriterium hallo2(kritU,*this,0);

  /*apply Lee criterium to the Equations -> find zero sectors*/
  lst unknown_vars;
  
  for (int j = 0; j < integralfamily.jule; j++)
    unknown_vars.append(unknown[j]);
  
  int count = 0;
  
  for (int sector_number = 0; sector_number < (1<<integralfamily.jule); sector_number++) {
    count = 0;
    
    for (int j = 0; j<integralfamily.jule; j++) {
      if ( (1<<j) & sector_number ){
        count++;
      }
    }
    
    unsigned forceTrivial;
    if(integralfamily.cutProps.size())
      forceTrivial = integralfamily.cutProps.size();
    else
      forceTrivial = 0;
    
    for (unsigned j = 0; j<integralfamily.cutProps.size(); j++) {
      if ( (1<<(integralfamily.cutProps[j]-1)) & sector_number ){
        forceTrivial--;
      }
    }
    
    if(forceTrivial){
      integralfamily.zeroSector.push_back(sector_number);
      load_bar(sector_number+1,(1<<integralfamily.jule),50,100);
      continue;
    }
    
    
    lst EQ3;
    hallo2.getList(sector_number,EQ3);
//     hallo2.print();
//     cout << EQ3 << endl;
    hallo.getList(sector_number,EQ3);
    ex solution = lsolve(EQ3,unknown_vars);
//     cout << EQ3 << endl;
//     cout << solution << endl;
    if (solution.nops() == 0) {
      numberOfNonTrivialSectors++;
      logger.set_level(2);
      logger<<"non trivial sector " << sector_number << " "<< count << "\n";
      logger.set_level(1);
      integralfamily.mask[count].push_back(sector_number);
    }
    else {
      integralfamily.zeroSector.push_back(sector_number);
    }
    load_bar(sector_number+1,(1<<integralfamily.jule),50,100);
  }
  
  /*Write results*/
  for(int i=0;i<integralfamily.jule+1;i++){
    std::sort(integralfamily.mask[i].begin(),integralfamily.mask[i].end());
    integralfamily.mask[i].resize(distance(integralfamily.mask[i].begin(),
			    unique(integralfamily.mask[i].begin(), integralfamily.mask[i].end())) );
  }
  ItVint endZero=unique(integralfamily.zeroSector.begin(), integralfamily.zeroSector.end());
  integralfamily.zeroSector.resize( distance(integralfamily.zeroSector.begin(),endZero) );
  sort(integralfamily.zeroSector.begin(),integralfamily.zeroSector.end());
  logger << "\nNon trivial sectors in total: " << numberOfNonTrivialSectors << "\n";

  ofstream fileZeroSector;
  fileZeroSector.open ((outputDir+"/sectormappings/"+integralfamily.name+"/trivialsector").c_str());
  
  for (unsigned it = 0, end1 = integralfamily.zeroSector.size(); it < end1; it++) {
    if (it< end1-1)
      fileZeroSector << integralfamily.zeroSector[it] << ",";
    else
      fileZeroSector << integralfamily.zeroSector[it];
  }
  
  fileZeroSector.close();
  
  ofstream nonTrivialSector;
  nonTrivialSector.open ((outputDir+"/sectormappings/"+integralfamily.name+"/nonTrivialSector").c_str());
  
  for(int i = 0; i < integralfamily.jule+1; i++){    
    for(unsigned it = 0; it < integralfamily.mask[i].size();it++){
      nonTrivialSector << integralfamily.mask[i][it] << " " << i << endl;
    }
  }
  
  nonTrivialSector.close();
  
  logger << "Trivial sectors in total: " << integralfamily.zeroSector.size()<<"\n";
  logger << "( " << clock.eval_time() << " s )\n\n";
  
  delete [] vec;
  delete [] feynman;
  delete [] unknown;
  delete [] exponent;
  
  if(numberOfNonTrivialSectors == 0){
    logger << "The system you try to reduce contains only zero integrals.\n";
    return 1;
  }
  else
    return 0;
  
}

int Kira::test(lst &matr){
  lst mZS;
  for (size_t i = 0; i < matr.nops(); i++){
    for (size_t j = 0; j < matr.nops(); j++){
      mZS.append(diff(integralfamily.loopVarList[i].subs(matr[i]),ex_to<symbol>(integralfamily.loopVarList[j])));
    }
  }
  matrix mZS2(integralfamily.loopVarList.nops(), integralfamily.loopVarList.nops(),mZS);
  
  int det = 0;
  det = something_int(mZS2.determinant());
  
  return det;
}

bool myfunction (const symmetries &i,const symmetries &j) { 
  
  if (i.topology<j.topology)
    return 1;
  if(i.topology>j.topology)
    return 0;
  
  if(i.sector < j.sector)
    return 1;
  if(i.sector > j.sector)
    return 0;
  
  return (i.symDOTS < j.symDOTS);
};

bool isEqual (const symmetries &i,const symmetries &j) {
  
  return (i.topology==j.topology && i.sector==j.sector && i.symDOTS == j.symDOTS);
  
}

class FUPolynom {
public:
  FUPolynom(ex listOfTerms,Kira& kira, int degree);
  ~FUPolynom();
  pair<ex,int**> **F;
  std::tuple<unsigned,unsigned,ex> **fSum;
  int *a;
  void setZero(int yy);
  void prepare();
  void recover();
  void remap(int i, int j);
  int compare(FUPolynom& comp);
  int compare3(FUPolynom& comp);
  int prepare_compare(FUPolynom& comp);
  void print();
  lst listOfTerms;
  vector<ex> * v_listOfTerms;
  int *fsize;
  int l_Indices;
  possymbol* feynman;
  unsigned nLoop;
  unsigned nOfPolynoms;
};

FUPolynom::FUPolynom(ex listOfTerms_, Kira & kira, int degree){
  
  for(unsigned i = 0; i < kira.invar.size(); i++){
    
    for (int j=-listOfTerms_.degree(1/kira.invar[i]); j<=listOfTerms_.degree(kira.invar[i]); ++j) {
      
      if(j!=0 && listOfTerms_.coeff(kira.invar[i],j)!=0){
        
	listOfTerms.append(listOfTerms_.coeff(kira.invar[i],j));
        
        listOfTerms_ = expand(listOfTerms_ - listOfTerms_.coeff(kira.invar[i],j)*(pow(kira.invar[i],j)));
      }
    }
  }
  if(listOfTerms_!=0)
    listOfTerms.append(listOfTerms_);
  
//   logger << "\n" << listOfTerms << "\n";
  
  nLoop = kira.integralfamily.loopVar.size()+degree;
  
  nOfPolynoms=listOfTerms.nops();
  
  v_listOfTerms = new vector<ex> [nOfPolynoms];
  
  l_Indices = kira.integralfamily.jule;
  
  a = new int [l_Indices];
  
  for(int i = 0; i < l_Indices; i++)
    a[i]=(1<<i);
  
  F = new pair<ex,int**> * [nOfPolynoms];
  fSum = new tuple<unsigned,unsigned,ex>* [nOfPolynoms];
  fsize = new int [nOfPolynoms];
    
  for (unsigned i = 0; i < nOfPolynoms; i++) {
    
    unsigned l_term = 0;
    if(is_a<add>(listOfTerms.op(i))){
      
      l_term = listOfTerms.op(i).nops();
      
      for( unsigned j = 0; j < l_term; j++ ) {
      
        v_listOfTerms[i].push_back(listOfTerms.op(i).op(j));
      }
    }
    else if( is_a<mul>(listOfTerms.op(i)) || is_a<symbol>(listOfTerms.op(i)) ){
      
      l_term = 1;
      
      v_listOfTerms[i].push_back(listOfTerms.op(i));
    }
    
    F[i]= new pair<ex,int**>  [l_term];
    fSum[i] = new tuple<unsigned,unsigned,ex> [l_term];
    
    for (unsigned j = 0; j < l_term; j++) {
      F[i][j].second = new int * [nLoop];
    }
  }
  
  
  feynman = new possymbol [l_Indices];
  generate_symbols(feynman,"b",l_Indices);
  
  
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for(unsigned j = 0; j < v_listOfTerms[i].size(); j++){
      
      int kk = 0;        
      
      F[i][j].first = 1;
      
      if(is_a<symbol>(v_listOfTerms[i][j])){
      
        for(int l = 0; l < l_Indices; l++){
          if(feynman[l]==v_listOfTerms[i][j]){
            F[i][j].second[kk++]=&a[l];
          }
        }
        continue;
      }
      
      
      if(is_a<power>(v_listOfTerms[i][j])) {
      
        exmap rpls;
        
        (v_listOfTerms[i][j]).match(pow(wild(0),wild(1)),rpls);
        
        for(int l = 0; l < l_Indices; l++){
        
          if(feynman[l]==rpls[wild(0)]){
          
            for(int ri = 0; ri < ex_to<numeric>(rpls[wild(1)]).to_int(); ri++)
            
              F[i][j].second[kk++]=&a[l];
          }
        }
        continue;
      }
      
      for(unsigned k = 0; k < v_listOfTerms[i][j].nops(); k++){
        
        if(k>nLoop){
          break;
        }
        if(is_a<power>(v_listOfTerms[i][j].op(k))) {

          exmap rpls;
          (v_listOfTerms[i][j].op(k)).match(pow(wild(0),wild(1)),rpls);
          
          for(int l = 0; l < l_Indices; l++){
            if(feynman[l]==rpls[wild(0)]){
              for(int ri = 0; ri < ex_to<numeric>(rpls[wild(1)]).to_int(); ri++)
                F[i][j].second[kk++]=&a[l];
            }
          }
        }
        else {
          for(int l = 0; l < l_Indices; l++){
            if(feynman[l]==v_listOfTerms[i][j].op(k)){
              F[i][j].second[kk++]=&a[l];
            }
          }
        }
      }
      if( (v_listOfTerms[i][j].nops()) > nLoop ){
        F[i][j].first = v_listOfTerms[i][j].op(v_listOfTerms[i][j].nops()-1);
      }
      else{
        F[i][j].first = 1;
      }
    }
  }
  prepare();
}

FUPolynom::~FUPolynom() {
  
  for (unsigned i = 0; i < nOfPolynoms; i++) {
    for (unsigned j = 0; j < v_listOfTerms[i].size(); j++) {
      delete [] F[i][j].second;
    }
    delete [] F[i];
    delete [] fSum[i];
  }
  delete [] v_listOfTerms;
  delete [] F;
  delete [] fSum;
  delete [] fsize;
  delete [] feynman;
  delete [] a;
  
};


void FUPolynom::prepare(){
  
  for(unsigned i = 0; i < nOfPolynoms; i++)
    fsize[i] = v_listOfTerms[i].size();
  
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for(unsigned j = 0; j < v_listOfTerms[i].size(); j++){
      get<0>(fSum[i][j])=0;
      get<1>(fSum[i][j])=0;
      for(unsigned k = 0; k < nLoop; k++){
	if(F[i][j].second[k][0]==0){
	  fsize[i]--;
	  get<0>(fSum[i][j])=0;
	  break;
	}
	get<0>(fSum[i][j])|=F[i][j].second[k][0];
        get<1>(fSum[i][j])+=F[i][j].second[k][0];
      }
      get<2>(fSum[i][j])=F[i][j].first;
    }
  }
  
  for(unsigned i = 0; i < nOfPolynoms; i++)
    sort(fSum[i],fSum[i]+v_listOfTerms[i].size(),dsc2);
  
}

void FUPolynom::setZero(int yy){
  
  a[yy]=0;
  
}

void FUPolynom::print(){

  for(unsigned i = 0; i < nOfPolynoms; i++){
    logger<< "size: " << fsize[i] << "\n";
    for (int j = 0; j < fsize[i]; j++){
      logger << get<0>(fSum[i][j]) << " " << get<1>(fSum[i][j]) << "*" << get<2>(fSum[i][j]) << " ";
    }
    logger << "\n";
  }
  
}

void FUPolynom::recover(){
  
  for(int i = 0; i < l_Indices; i++)
    a[i]=(1<<i);
  for(unsigned i = 0; i < nOfPolynoms; i++)
    fsize[i]=v_listOfTerms[i].size();
  
}

void FUPolynom::remap(int i, int j){

  a[i]=(1<<j);
  
}

int FUPolynom::prepare_compare(FUPolynom& comp){
  
//   Clock clock;
  
  for(unsigned i = 0; i < nOfPolynoms; i++)
    fsize[i] = v_listOfTerms[i].size();
  
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for(unsigned j = 0; j < v_listOfTerms[i].size(); j++){
      get<0>(fSum[i][j])=0;
      get<1>(fSum[i][j])=0;
      for(unsigned k = 0; k < nLoop; k++){
	if(F[i][j].second[k][0]==0){
	  fsize[i]--;
	  get<0>(fSum[i][j])=0;
	  break;
	}
	get<0>(fSum[i][j]) |= F[i][j].second[k][0];
        get<1>(fSum[i][j]) += F[i][j].second[k][0];
      }
      get<2>(fSum[i][j]) = F[i][j].first;
    }
    
    if (fsize[i] != comp.fsize[i]){
//       logger << "return 2 ( " << clock.eval_time() << " s )\n";  
      return 1;
    }
    
    sort(fSum[i],fSum[i]+v_listOfTerms[i].size(),dsc2);
    
    for (int j = 0; j < fsize[i]; j++) {
      if(get<0>(fSum[i][j]) != get<0>(comp.fSum[i][j]) && get<1>(fSum[i][j]) != get<1>(comp.fSum[i][j])){
// 	logger << "return 2 ( " << clock.eval_time() << " s )\n";  
	return 2;
      }
    }
  }
//   logger << "( " << clock.eval_time() << " s )\n";  
  return 0;
}


int FUPolynom::compare(FUPolynom& comp){

  for(unsigned i = 0; i < nOfPolynoms; i++){
    for (int j = 0; j < fsize[i]; j++) {
      if(get<0>(fSum[i][j]) != get<0>(comp.fSum[i][j]) && get<1>(fSum[i][j]) != get<1>(comp.fSum[i][j]))
	return 0;
    }
  }
  
  for(unsigned i = 0; i < nOfPolynoms; i++){
    for (int j = 0; j < fsize[i]; j++) {
      if(get<2>(fSum[i][j]) != get<2>(comp.fSum[i][j]))
	return 0;
    }
  }
  
  return 1;
  
}  

int FUPolynom::compare3(FUPolynom& comp){
  
  for(unsigned i = 0; i < nOfPolynoms; i++){
    if (fsize[i] != comp.fsize[i])
      return 0;
  }
  
  return 1;
  
}


int Kira::prepare_symmetry(){
  
//   for (int j=0;j<integralfamily.biggestBound+1;++j){
//     for (size_t itt=0; itt!=integralfamily.mask[j].size();++itt){ // all possible non trivial sectors
//       for(size_t itS=0; itS!=integralfamily.sector2Reduce.size(); itS++) {
// 	if((integralfamily.sector2Reduce[itS] & integralfamily.mask[j][itt]) == integralfamily.mask[j][itt]) {
// 	  integralfamily.allowSector[j].push_back(integralfamily.mask[j][itt]);
// 	  break;
// 	}
//       }
//     }
//   }
  
  for (int j=0;j<integralfamily.biggestBound+1;++j){
    for (size_t itt=0; itt!=integralfamily.mask[j].size();++itt){ // all possible non trivial sectors
      
      int sectorNumber=integralfamily.mask[j][itt];
      int num_ones=0;
      unsigned testSector = sectorNumber;
      for(size_t i = 0; i < SEEDSIZE; ++i, testSector >>= 1) {
	if ((testSector & 1) == 1)
	  ++num_ones;
      }
      if(num_ones < SEEDSIZE){
	int flagy = 0;
	for(size_t it=0;it<integralfamily.lowestSectors.size();it++){
	  if((integralfamily.lowestSectors[it] & sectorNumber) == integralfamily.lowestSectors[it] ){
	    flagy = 1;
	    break;
	  }
	}
	if(!flagy){
	  integralfamily.lowestSectors.push_back(sectorNumber);
	}
      }
    }
  }
  
  if(file_exists((outputDir+"/sectormappings/"+integralfamily.name+"/magicFlag").c_str())){
    ifstream input;
    input.open((outputDir+"/sectormappings/"+integralfamily.name+"/magicFlag").c_str());
    
    string magicFlag;
    
    input >> magicFlag;
    if(integralfamily.magic_relations != magicFlag){
      logger << "You have ones generated the symmetries with the\n";
      logger << "option magic_relations switched "; 
      if(magicFlag == "true")
	logger << "on.\n";
      else
	logger << "off.\n";
      
      logger << "If you change the option magic_relations compared\n";
      logger << "to the previous run you need to delete the directory:\n";
      logger << (outputDir+"/sectormappings/"+integralfamily.name) << "\n";
      exit(0);
    }
  }
  else{
    ofstream output;
    output.open((outputDir+"/sectormappings/"+integralfamily.name+"/magicFlag").c_str());
    
    string magicFlag;
    
    output << integralfamily.magic_relations;
  }
  
  int flagSYM = read_symmetries((outputDir+"/sectormappings/"+integralfamily.name+"/sectorRelations"), integralfamily.symVec);
  int flagREL = read_symmetries((outputDir+"/sectormappings/"+integralfamily.name+"/sectorSymmetries"), integralfamily.relVec);

  if(flagSYM && flagREL)
    return 1;
  return 0;
}

// class ThreadPool;
//   
// // our worker thread objects
// class Worker {
// public:
//     Worker(ThreadPool &s) : pool(s) { }
//     void operator()();
// private:
//     friend class Kira;
//     ThreadPool &pool;
// };
//   
// // the actual thread pool
// class ThreadPool {
// public:
//     ThreadPool()  :  stop(false){};
//     void initialize(uint32_t);
//     template<class F>
//     void enqueue(const F &f);
//     ~ThreadPool();
// private:
//     friend class Worker;
//     friend class Kira;
//  
//     // need to keep track of threads so we can join them
//     std::vector<std::thread > workers;
//  
//     // the task queue
//     std::deque<std::function<void()> > tasks;
//  
//     // synchronization
//     std::mutex queue_mutex;
//     std::condition_variable condition;
//     bool stop;
// };
// 
// 
// // add new work item to the pool
// template<class F>
// void ThreadPool::enqueue(const F &f)
// {
//     { // acquire lock
//         std::unique_lock<std::mutex> lock(queue_mutex);
//          
//         // add the task
//         tasks.push_back(std::function<void()>(f));
//     } // release lock
//      
//     // wake up one thread
//     condition.notify_one();
// }

// ThreadPool poolS;
// std::mutex mutexSym;

int Kira::symmetry_finder(int it, int itt, string topoName,
			   int j, int array[], int array2[], int array3[], int flag){
  symmetries halloB;
  
  halloB.symDOTS = 0;
  
  for (int g=0;g<integralfamily.jule;g++){
    halloB.ing[g]=-1;
  }
  for (int g=0;g<j;g++){
    halloB.ing[array2[g]]=array3[array[g]];
  }
  
  
  unsigned externalSym = 1;
//   if(integralfamily.name == topoName)
//     externalSym=topology[topoName].propsMomFlowB.size();
  externalSym = integralfamily.propsMomFlowB.size();
//   poolS.initialize(8);
  for (unsigned klop = 0; klop < externalSym; klop++){
    
//     uint32_t totalJ = (1<<j);
//     uint32_t countJ = 0;
//     std::mutex m;
//     std::condition_variable cond_var;
//     bool processed = false;
    
//     if(totalJ != 0){
      
      for (int gi = 0; gi < (1<<j); gi++){
	
// 	poolS.enqueue([this, &j, &halloB, &array, &array2, &topoName, gi, &klop, &array3, &it, &itt, &flag ,&totalJ, &countJ, &processed, &cond_var, &m]()
// 	{
	  int re;
	  lst hastosolve;
	  halloB.det = 0;
	  
	  int skipDo = 0;
	  
	  for (int g = 0; g < j; g++){
	    if ( !(!(1<<array[g] & integralfamily.maskCut) == !(1<<array2[g] & topology[topoName].maskCut))  ){
	      skipDo = 1;
	      break;
	    }
	    
	    if( (gi & (1<<g)) ){
	      hastosolve.append( -(integralfamily.propsMomFlowA[array2[g]])
	      == (topology[topoName].propsMomFlowB[klop][array3[array[g]]]));
	    }
	    if( !(gi & (1<<g)) ){
	      hastosolve.append( (integralfamily.propsMomFlowA[array2[g]])
	      == (topology[topoName].propsMomFlowB[klop][array3[array[g]]]));
	    }
	  }
	  
          if(skipDo)
            continue;
	  
    //       logger << hastosolve << "\n";
	  
	  halloB.subst.remove_all();
	  lst solution;
	  
    //       logger << topology[topoName].loopVarList2 << "\n";
    //       logger << topology[topoName].loop2loop << "\n";
	  
	  solution.append((lsolve(hastosolve,topology[topoName].loopVarList).subs(topology[topoName].loop2loop)));
	  
	  
	  for(size_t hi = 0; hi < solution[0].nops(); hi++){
	    halloB.subst.append(solution[0][hi]);
	  }
	  
	  
	  if( solution[0].nops()){
	    re = test(halloB.subst);
	    
    //         logger << halloB.subst << "\n";
    //         logger << hastosolve << "\n";
    //         logger << re << "\n";
	    
	    if(re!=0){
	      
	      halloB.det = re;
	      halloB.sector = topology[topoName].mask[j][itt];
	      halloB.nOfProps = j;
	      halloB.topology = topology[topoName].topology;
	      halloB.externalSymmetry=klop;
	      
// 	      std::unique_lock<std::mutex> lckCritical (mutexSym,std::defer_lock);
// 	      // critical section 
// 	      lckCritical.lock();
// 	      {
		logger.set_level(2);
		if(get<2>(externalTransf[klop]) != controlSymmetries && klop != 0){
		  halloB.symDOTS = 1;
		  logger <<"Only for dots: ";
		}
		if(flag){
		  logger <<"Kira maps this sector " << integralfamily.mask[j][it] << " to this sector " << halloB.sector << "\n";
		  logger <<"with external Symmetry " << klop << "\n";
		  integralfamily.symVec[integralfamily.mask[j][it]].push_back(halloB);
		  integralfamily.skipSector[j].push_back(integralfamily.mask[j][it]);
		}
		else{
		  logger <<"Kira maps this sector " << integralfamily.mask[j][it] << " to this sector " << halloB.sector << "\n";
		  logger <<"with external Symmetry " << klop << "\n";
		  integralfamily.relVec[integralfamily.mask[j][it]].push_back(halloB);
		}
	  topology[topoName].sectorWsymmetry[halloB.nOfProps].insert(halloB.sector);
	  
		logger.set_level(1);
		
// 	      }  
// 	      lckCritical.unlock();
	      return 1;
	    }
	    else{
	      logger.set_level(2);
	      logger << "fail: " << re << "\n";
	      logger.set_level(1);
	      halloB.det = 0;
	      halloB.subst.remove_all();
	    }
	  }
// 	  {
// 	    std::lock_guard<std::mutex> lock(m);
// 	    countJ++;
// 	    if(countJ == totalJ){
// 	      processed = true;
// 	      cond_var.notify_one();
// 	    }
// 	  }
// 	
// 	});
//       }
//       {
// 	std::unique_lock<std::mutex> lock(m);
// 	cond_var.wait(lock, [&processed]{return processed;});
//       }
    }
	    
  }
  return 0;
}

int Kira::skip_symmetry(string topoName,int j,int it,int itt) {
  
  int skipTopSym = 1;
  
  for(vector<int>::iterator topIt = topology[topoName].topLevelSectors.begin(); topIt != topology[topoName].topLevelSectors.end(); topIt++){
    
    if( (topology[topoName].mask[j][itt] & (*topIt)) == topology[topoName].mask[j][itt] ){
      
      skipTopSym = 0;
    }
  }
  
  if(skipTopSym == 0){
    
    skipTopSym = 1;
    for(auto topIt = integralfamily.topLevelSectors.begin(); topIt != integralfamily.topLevelSectors.end(); topIt++){
      if( (integralfamily.mask[j][it] & (*topIt)) == integralfamily.mask[j][it]){
	
	skipTopSym = 0;
      }
    }
  }
  
  if(topology[topoName].magic_relations=="true"){
    
    for(vector<int>::iterator topIt = topology[topoName].topLevelSectors.begin(); topIt != topology[topoName].topLevelSectors.end(); topIt++){
      if( ((*topIt) & topology[topoName].mask[j][itt]) 
// 	== topology[topoName].mask[j][itt]
// 	&& integralfamily.name == topoName
      ){
	
	skipTopSym = 0;
      }
    }
  }
  
  if(skipTopSym == 1)
    return 1;
  
//   if(j>0){
//     
//     {
//       int countPos=0;
//     
//       for(auto itSet: topology[topoName].sectorWsymmetry[j-1]){
// 	if((topology[topoName].mask[j][itt] & itSet) == itSet){
// 	  countPos++;
// 	}
//       }
//       if(countPos==0){
// 	skipTopSym = 1;
//       }
//       else{
// 	skipTopSym = 0;
//       }
//     }
//     if(skipTopSym == 0){
//       int countPos = 0;
//       for(auto itSet: integralfamily.sectorWsymmetry[j-1]){
// 	if((integralfamily.mask[j][it] & itSet) == itSet){
// 	  countPos++;
// 	}
//       }
//       if(countPos==0){
// 	skipTopSym = 1;
//       }
//       else{
// 	skipTopSym = 0;
//       }
//     }
//     
//   }
  
  vector<unsigned>::iterator itLowestSKIP = 
	      find(integralfamily.lowestSectors.begin(), 
		  integralfamily.lowestSectors.end(), integralfamily.mask[j][it]);
	      
  if (itLowestSKIP != integralfamily.lowestSectors.end())
    skipTopSym = 0;
  
  if(skipTopSym == 1)
    return 1;
  
  if( integralfamily.mask[j][it] == topology[topoName].mask[j][itt] 
    && integralfamily.name == topoName)
    return 0;
  
  
  vector<int>::iterator itSKIP = find (integralfamily.skipSector[j].begin(), 
				      integralfamily.skipSector[j].end(), integralfamily.mask[j][it]);
  if (itSKIP != integralfamily.skipSector[j].end()){
    return 1;
  }
  
  
  if(j>0 && integralfamily.skipSector[j-1].size()){
    int skip=1;
    for(unsigned skipIT = 0; skipIT < integralfamily.skipSector[j-1].size(); skipIT++){
      if ( (integralfamily.skipSector[j-1][skipIT] & integralfamily.mask[j][it]) == integralfamily.skipSector[j-1][skipIT]){
	skip=0;
	break;
      }
    }
    if(skip==1){
      return 1;
    }
  }
  
  return 0;
  
}

void Kira::symmetry_relations(string topoName) {
  
  FUPolynom fpolynom1(topology[topoName].F,*this,1);
  FUPolynom upolynom1(topology[topoName].U,*this,0);
  FUPolynom fpolynom2(integralfamily.F,*this,1);
  FUPolynom upolynom2(integralfamily.U,*this,0);
  
  Clock clock;
  
  possymbol* feynman = new possymbol [integralfamily.jule];
  generate_symbols(feynman,"b",integralfamily.jule);
  int array[integralfamily.jule], array2[integralfamily.jule], array3[integralfamily.jule];
//   int arrayBonus[integralfamily.jule];
  
  for (int i = 0; i < integralfamily.jule; i++){
    array[i] = i;
  }
  
  for (int j = 0; j < integralfamily.biggestBound+1; ++j){
    
    for (size_t itt = 0; itt < topology[topoName].mask[j].size(); ++itt) { // all possible non trivial sectors
      
      if(integralfamily.mask[j].size()==0)
	continue;
      
//       upolynom1.recover();
//       fpolynom1.recover();
//       
//       int comB=0;
//       int k = 0;
//       for (int gg = 0; gg < integralfamily.biggestBound; gg++)
//         array3[gg] = -1;
//       for (int y = 0; y<integralfamily.jule; y++) {
//         if (((1<<y) & topology[topoName].mask[j][itt])){
//           array3[k++] = y;
// 	  comB += topology[topoName].invarID[y]*powerINT(100,topology[topoName].invarID[y]);
//         }
//         else{
// 	  upolynom1.setZero(y);
// 	  fpolynom1.setZero(y);
//         }
//       }
//       upolynom1.prepare();
//       fpolynom1.prepare();
      
      size_t it2 = 0;
      
      if(integralfamily.name==topoName)
	it2 = itt;
      
      int constructOnce = 0;
      int comB = 0;
      
      for (size_t it = it2; it < integralfamily.mask[j].size(); it++) {
// 	logger << "start: "<< integralfamily.mask[j][it] << " " << topology[topoName].mask[j][itt] << "\n";
	
	//map itt -> it
	if(skip_symmetry(topoName,j,it,itt) == 1) {
	  continue;
        }
        
        if(constructOnce==0){
	  
	  upolynom1.recover();
	  fpolynom1.recover();
	  
	  int k = 0;
	  for (int gg = 0; gg < integralfamily.biggestBound; gg++)
	    array3[gg] = -1;
	  for (int y = 0; y<integralfamily.jule; y++) {
	    if (((1<<y) & topology[topoName].mask[j][itt])){
	      array3[k++] = y;
	      comB += topology[topoName].invarID[y]*powerINT(100,topology[topoName].invarID[y]);
	    }
	    else{
	      upolynom1.setZero(y);
	      fpolynom1.setZero(y);
	    }
	  }
	  upolynom1.prepare();
	  fpolynom1.prepare();
	  constructOnce++;
	}
        
	
	upolynom2.recover();
	fpolynom2.recover();
        
	for (int g=0;g<integralfamily.biggestBound;g++)
          array2[g]=-1;
        
	int k = 0;
        int comA=0;
	for (int y = 0; y<integralfamily.jule; y++) {
          if (((1<<y) & integralfamily.mask[j][it])){
            array2[k++] = y;
	    comA+=integralfamily.invarID[y]*powerINT(100,integralfamily.invarID[y]);
          }
          else{
	    upolynom2.setZero(y);
	    fpolynom2.setZero(y);
          }
        }
//         cout << comA << " " << comB << endl;
	if(comA != comB)
	  continue;

        
	int count=0;
	
// 	int foundSym = 1;
//         logger << "start: "<< integralfamily.mask[j][it] << " " << topology[topoName].mask[j][itt] << "\n";
	std::sort(array, array+(integralfamily.jule-1));
	do {
	  int skipPERM = 0;
          for (int g=0;g<j;g++){
	    
            upolynom2.remap(array2[g],array3[array[g]]);
	    fpolynom2.remap(array2[g],array3[array[g]]);
            
            if(integralfamily.invarID[array2[g]] != topology[topoName].invarID[array3[array[g]]]){
	      skipPERM = 1;
	      break;
	    }
          }
          if(skipPERM==1)
	    continue;
	  
	  
// 	  logger<<"U-Polynom\n";
          upolynom2.prepare();
	    
// 	  upolynom1.print();
// 	  upolynom2.print();
	
	  
// 	  if(!upolynom1.compare4(upolynom2))
// 	    break;
	  
// 	  int slag = upolynom2.prepare_compare(upolynom1);
// 	  if(slag==1)
// 	    break;mask
// 	  if(slag==2)
// 	    continue;

	  if(!upolynom1.compare3(upolynom2)){
	    break;
          }
	  
	  if(!upolynom1.compare(upolynom2) && count > 0){
	    continue;
	  }
	  
// 	  logger<<"F-Polynom\n";
// 	  int slagF = fpolynom2.prepare_compare(fpolynom1);
// 	  if(slagF==1)
// 	    break;
// 	  if(slagF==2){
// 	    continue;
// 	  }
	  
	  fpolynom2.prepare();
// 	  fpolynom1.print();
// 	  fpolynom2.print();
	  if(!fpolynom1.compare3(fpolynom2)){
	    break;
          }
	  
	  if(!fpolynom1.compare(fpolynom2))
	    continue;
	  
//           for (int i = 0; i < integralfamily.jule; i++){
//             arrayBonus[i] = array[i];
//           }
          
          
// 	  int skipSYM = 0;
	  
	  if(integralfamily.name==topoName){
	    if(integralfamily.mask[j][it] != topology[topoName].mask[j][itt]){
//               foundSym = 0;
	      /*skipSYM = */symmetry_finder(it, itt, topoName, j, array, array2, array3,1);
//               if (skipSYM == 1){
// //                 foundSym = 1;
//                 break;
//               }
	    }
	    else if(count && integralfamily.mask[j][it] == integralfamily.mask[j][itt]){
// 	      foundSym = 0;
              /*skipSYM = */symmetry_finder(it, itt, topoName, j, array, array2, array3,0);
//               if (skipSYM == 1){
// //                 foundSym = 1;
//                 break;
//               }
	    }
	  }
	  else{
// 	    foundSym = 0;
            /*skipSYM = */symmetry_finder(it, itt, topoName, j, array, array2, array3,1);
//             if (skipSYM == 1){
// //                 foundSym = 1;
//                 break;
//             }
	  }
	  
// 	  if(foundSym==0){
//             
//             symmetries halloB;
//             
//             halloB.symDOTS = 1;
//             
//             for (int g=0;g<integralfamily.jule;g++){
//               halloB.ing[g]=-1;
//             }
//             
//             for (int g=0;g<j;g++){
//               halloB.ing[array2[g]]=array3[arrayBonus[g]];
//             }
//             
//             halloB.det = 1;
//             
//             for(unsigned it = 0; it < integralfamily.loopVar.size(); it++){
//           
//               ex substleft = 0;
//               ex substright = 0;
//               halloB.subst.append(substleft==substright);
//             }
//             halloB.sector = topology[topoName].mask[j][itt];
//             halloB.nOfProps = j;
//             halloB.topology = topology[topoName].topology;
//             halloB.externalSymmetry=0;
//             
//             if(integralfamily.name==topoName){
//               if(integralfamily.mask[j][it] != topology[topoName].mask[j][itt]){
//                 integralfamily.symVec[integralfamily.mask[j][it]].push_back(halloB);
//                 logger.set_level(2);
//                 logger <<"Kira maps this sector " << integralfamily.mask[j][it] << " to this sector " << halloB.sector << " (only dots)\n";
//                 logger.set_level(1);
//                 break;
//               }
//               else if(count && integralfamily.mask[j][it] == integralfamily.mask[j][itt]){
//                 integralfamily.relVec[integralfamily.mask[j][it]].push_back(halloB);
//                 logger.set_level(2);
//                 logger <<"Kira maps this sector " << integralfamily.mask[j][it] << " to this sector " << halloB.sector << " (only dots)\n";
//                 logger.set_level(1);
//                 break;
//               }
//             }
//             else{
//               integralfamily.symVec[integralfamily.mask[j][it]].push_back(halloB);
//               logger.set_level(2);
//               logger <<"Kira maps this sector " << integralfamily.mask[j][it] << " to this sector " << halloB.sector << " (only dots)\n";
//               logger.set_level(1);
//               break;
//             }
//           }
          count++;
        } while ( next_permutation( array, array+j) );
      }
    }
    
    sort(integralfamily.skipSector[j].begin(),integralfamily.skipSector[j].end());
    vector<int>::iterator itSKIP;
    itSKIP = std::unique (integralfamily.skipSector[j].begin(),integralfamily.skipSector[j].end());  
    integralfamily.skipSector[j].resize( std::distance(integralfamily.skipSector[j].begin(),itSKIP) );    
//     load_bar(j+1,integralfamily.biggestBound+1,50,100);
  }
  
  int countR = 0;
  for(int it = 0; it < (1<<integralfamily.jule)+1; it++) {
    sort(integralfamily.symVec[it].begin(),integralfamily.symVec[it].end(),myfunction);
//     vector<symmetries>::iterator itSKIP;
//     itSKIP = std::unique (integralfamily.symVec[it].begin(),integralfamily.symVec[it].end(),isEqual);  
//     integralfamily.symVec[it].resize( std::distance(integralfamily.symVec[it].begin(),itSKIP) );
    countR+=integralfamily.symVec[it].size();
  }

  int countS = 0;
  for(int it = 0; it < (1<<integralfamily.jule)+1; it++) {
    sort(integralfamily.relVec[it].begin(),integralfamily.relVec[it].end(),myfunction);
//     vector<symmetries>::iterator itSKIP;
//     itSKIP = std::unique (integralfamily.relVec[it].begin(),integralfamily.relVec[it].end(),isEqual);  
//     integralfamily.relVec[it].resize( std::distance(integralfamily.relVec[it].begin(),itSKIP) );
    countS+=integralfamily.relVec[it].size();
  }

  write_symmetries((outputDir+"/sectormappings/"+integralfamily.name+"/sectorRelations"),integralfamily.symVec);  
  write_symmetries((outputDir+"/sectormappings/"+integralfamily.name+"/sectorSymmetries"),integralfamily.relVec);
  
  logger << "\nSector relations: " << countR << "\n";
  logger << "Sector symmetries: " << countS << "\n";
  
  delete [] feynman;
  
  logger << "( " << clock.eval_time() << " s )\n";  
}

void Kira::write_symmetries(const string otputName,SYM symVec[]){
  
  ofstream fileSymVec;
  fileSymVec.open (otputName.c_str());
  
  for(int i = 0; i < (1<<integralfamily.jule)+1; i++) {
    for (unsigned it = 0; it < symVec[i].size();it++){
      symmetries halloB = symVec[i][it];
      fileSymVec << i << " ";
      
      for(unsigned itt = 0; itt < halloB.subst.nops(); itt++){
	fileSymVec << halloB.subst.op(itt).lhs() << " " << halloB.subst.op(itt).rhs() << " ";
      }
      fileSymVec << halloB.det << " ";
      fileSymVec << halloB.sector << " ";
      fileSymVec << halloB.nOfProps << " ";
      fileSymVec << halloB.externalSymmetry << " ";
      fileSymVec << get<0>(externalTransf[halloB.externalSymmetry]) << " ";
      fileSymVec << get<1>(externalTransf[halloB.externalSymmetry]) << " ";
      for (int g = 0; g < integralfamily.jule; g++){
	fileSymVec <<halloB.ing[g] << " ";
      }
      fileSymVec << halloB.symDOTS << " ";
      fileSymVec << halloB.topology << endl;
    }
  }
  
  fileSymVec.close();
}

int Kira::read_symmetries(const string inputName, SYM symVec[]){
  
  ifstream input;
  int count=0;
  if(file_exists(inputName.c_str())){
    input.open(inputName.c_str());
    logger << "read from: " << inputName << "\n";
    parser symbolReader(GiNaCSymbols);
    while(1){

      symmetries halloB;
      
      int sector;
      string skipRead;
      
      if(!(input >> sector)) break;
      for(unsigned it = 0; it < integralfamily.loopVar.size(); it++){
	
	string substleft;
	string substright;
	
	if(!(input >> substleft)) break;
	if(!(input >> substright)) break;
	halloB.subst.append(symbolReader(substleft)==symbolReader(substright));
      }
      if(!(input >> halloB.det)) break;
      if(!(input >> halloB.sector)) break;
      if(!(input >> halloB.nOfProps)) break;
      if(!(input >> halloB.externalSymmetry)) break;
      if(!(input >> skipRead)) break;
      if(!(input >> skipRead)) break;
      for (int g = 0; g < integralfamily.jule; g++){
	if(!(input >> halloB.ing[g])) break;
      }
      if(!(input >> halloB.symDOTS)) break;
      if(!(input >> halloB.topology)) break;
      
      symVec[sector].push_back(halloB);
      
      if(sector!=halloB.sector){
	integralfamily.skipSector[halloB.nOfProps].push_back(sector);
      }
      count++;
    }
    logger <<"Number of symmetries read out: " << count << "\n";
    input.close();
    
    return 1;
  }
  return 0;
}
